# CREDITS: this example is adapted from the pygame website
#          https://www.pygame.org/docs/tut/PygameIntro.html

import random
import pygame


WIDTH = 640
HEIGHT = 480

PADDLE_WIDTH = int(.25 * WIDTH)
PADDLE_HEIGHT = 30

FPS = 60

pygame.init()

BLACK = (0, 0, 0)
BLUE = (0, 0, 255)


class Ball:
    """
    A class used to represent a beachball in pygame

    Attributes
    ----------
    speed_x (int or float): the speed in the x direction
    speed_y (int or float): the speed in the y direction
    mega_hard (bool): whether or not we are playing in "mega-hard" mode
    image (pygame.surface.Surface): a pygame representation of the beachball image
    ball_rect (pygame.rect.Rect): a pygame rect object representing the position of the ball

    Methods
    -------
    update(paddle)
        updates the speed of the ball given the current position.
        Returns 1 if the ball hit the paddle, -1 if it hit the ground, and 0 otherwise.

    render(screen)
        Renders the ball image on the screen
    """
    def __init__(self, speed, mega_hard):
        """
        Initializes a Ball object, loads the image

        Args:
            speed (int or float): the initial speed of the ball
            mega_hard (bool): whether the ball's speed should be updated using the "mega-hard" mode
        """
        self.speed_x = speed
        self.speed_y = speed
        self.mega_hard = mega_hard
        self.image = pygame.image.load("intro_ball.gif")
        # make the beachball smaller
        self.image = pygame.transform.scale(self.image, (50, 50))
        self.ball_rect = self.image.get_rect()
        
    def update(self, paddle):
        """
        Updates the speed of the ball given the current position. Also determines whether the game
        should end or if the player should get a point.
        Changes the color of the paddle if the ball collided with it.

        Args:
            paddle (Paddle): the paddle in the game

        Returns:
            int: returns 1 if the ball hit the paddle, -1 if it hit the ground, and 0 otherwise.
        """
        self.ball_rect = self.ball_rect.move((self.speed_x, self.speed_y))
        if self.ball_rect.left < 0 or self.ball_rect.right > WIDTH:
            self.speed_x *= -1
        if self.ball_rect.top < 0: # removed check for height
            self.speed_y *= -1
        # changes from older version below...
        if self.ball_rect.colliderect(paddle.paddle_rect):
            self.speed_y *= -1
            # if we collided, make the paddle a new color
            paddle.change_color()
            
            # if this is mega-hard, we should increase the speed
            if self.mega_hard:
                self.speed_x *= 1.25
                self.speed_y *= 1.25
            return 1
        if self.ball_rect.bottom > HEIGHT:
            return -1
        return 0
            
    def render(self, screen):
        """
        Renders the beachball on it's current position on the screen

        Args:
            screen (pygame.surface.Surface): the screen object

        Returns: None
        """
        screen.blit(self.image, self.ball_rect.topleft)


class Paddle:
    """
    A class used to represent a paddle in pygame

    Attributes
    ----------
    paddle_rect (pygame.rect.Rect): a pygame rect object representing the position of the paddle
    move_pixels (int): the number of pixels to move by when keys are pressed
    color (Tuple[int, int, int]): the color of the paddle
        Assumptions: all of the ints are between 0 and 255

    Methods
    -------
    move(right)
        moves the Paddle to the left or right
    render(screen)
        Renders the paddle rectangle on the screen
    change_color()
        Makes self.color a new random color
    """
    def __init__(self):
        """
        Initializes a Paddle object

        Args: N/A
        """
        self.paddle_rect = pygame.Rect(0, HEIGHT - PADDLE_HEIGHT, PADDLE_WIDTH, PADDLE_HEIGHT)
        self.move_pixels = 50
        self.color = BLUE
        
    def move(self, right):
        """
        Updates the rect's x position

        Args:
            right (bool): True if we should move to the right, False if we should move to the left

        Returns: None
        """
        if right:
            self.paddle_rect.x += self.move_pixels
        else:
            self.paddle_rect.x -= self.move_pixels
        
    def render(self, screen):
        """
        Renders the paddle on it's current position on the screen

        Args:
            screen (pygame.surface.Surface): the screen object

        Returns: None
        """
        pygame.draw.rect(screen, self.color, self.paddle_rect)
        
    def change_color(self):
        """
        Updates the paddle's color to a random color

        Args: N/A

        Returns: None
        """
        self.color = (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))
        
        


def play_game():
    """
    Plays the game, keeping the beachball in the air with the paddle
    Some of the extensions are implemented

    Arguments: N/A
    
    Returns: None
    """
    screen = pygame.display.set_mode((WIDTH, HEIGHT))
    clock = pygame.time.Clock()
    
    # play around with these settings for difficulty
    speed = 2
    mega_hard = True
    
        
    ball = Ball(speed, mega_hard)
    paddle = Paddle()
    
    # keep track of points
    points = 0
        
    # Main game loop
    while True:

        # make the background black again
        screen.fill(BLACK)

        # Obtain any user inputs
        event = pygame.event.poll()
        if event.type == pygame.QUIT:
          break
        if event.type == pygame.KEYDOWN:
            if event.key == pygame.K_RIGHT:
                paddle.move(True)
            elif event.key == pygame.K_LEFT:
                paddle.move(False)        

        # update the position/direction of the ball
        state = ball.update(paddle)
        
        # check the state: add a point for 1, end game for -1
        if state == 1:
            points += 1
            print("Current score:", points)
        elif state == -1:
            break
        
        # render the paddle and the ball
        ball.render(screen)
        paddle.render(screen)

        
        # Render next frame
        pygame.display.update()
        clock.tick(FPS)
        
    print("Game ended with", points, "points")

        
if __name__ == "__main__":
    play_game()
